import os, flask, flask_cors

from ..api import API
from ..overloads import overloads
from .editor import get_editor_bp
from .player import get_player_bp


def create_app(api: API):
    app = flask.Flask(__name__)
    flask_cors.CORS(app)

    editor = get_editor_bp(api)
    app.register_blueprint(editor, url_prefix="/_editor")

    player = get_player_bp(api)
    app.register_blueprint(player)

    return app


def serve(workspace_root: str, port: int, debug, use_reloader):
    os.environ["ABSTRA_SERVER"] = "true"
    overloads()

    api = API(workspace_root)
    app = create_app(api)
    app.run(host="localhost", port=port, debug=debug, use_reloader=use_reloader)



import sys
from ..resources.resources import Resource
from .. import messages
from .. import utils
from .. import apis


class Vars(Resource):
    @staticmethod
    def list():
        vars = apis.list_workspace_vars()
        messages.print_vars(vars)

    @staticmethod
    def add(*args, **kwargs):
        vars = list(args)
        file = kwargs.get("f") or kwargs.get("file")
        if file:
            with open(file, "r") as f:
                vars.extend([v for v in f.read().split("\n") if v])

        if len(vars) == 0:
            messages.no_variables()
            return

        processed_vars = []
        processed_names = []
        for var in vars:
            name, value = utils.parse_env_var(var)
            if not name or not value:
                messages.invalid_variable(var)
                sys.exit(1)
            if name in processed_names:
                messages.duplicate_variable(var)
                sys.exit(1)
            processed_vars.append({"name": name, "value": value})

        added_vars = apis.add_workspace_vars(processed_vars)
        messages.print_vars(added_vars)
        print(f"\nAdded {len(added_vars)} environment variables")

    @staticmethod
    def remove(*args, **kwargs):
        deleted_vars = apis.delete_workspace_vars(args)
        messages.print_vars(deleted_vars)
        print(f"\nDeleted {len(deleted_vars)} vars")



from .sqlite import *
from .legacy import *  # deprecated
import os


class Table:
    def __init__(self, db: SqliteDB, table_name: str) -> None:
        self.db = db
        self.table_name = table_name

    def select(self, where: str, rows: list[str] = ["*"], **params) -> list:
        return self.db.select(self.table_name, where, rows, **params)

    def insert(self, values: dict[str, Any] = {}) -> None:
        return self.db.insert(self.table_name, values)

    def update(self, where: str, set: dict[str, Any] = {}, **params) -> None:
        return self.db.update(self.table_name, where, set, **params)

    def delete(self, where: str, **params) -> None:
        return self.db.delete(self.table_name, where, **params)


def get_db():
    db_url = os.getenv("ABSTRA_DATABASE_URL")
    if db_url.startswith("postgres://"):
        raise NotImplementedError("Postgres is not supported yet")
    elif db_url.startswith("mysql://"):
        raise NotImplementedError("MySQL is not supported yet")
    else:
        return SqliteDB(db_url)


def get_table(table_name: str):
    return Table(get_db(), table_name)



import json

import abstra.hooks as hooks


class TestUtils:
    def test_get_request_without_file(self):
        body, query, headers = hooks.get_request()
        assert body is None
        assert query is None
        assert headers is None

    def test_get_request_with_local_file(self):
        local_file = "/tmp/request.json"
        with open(local_file, "w") as f:
            f.write(
                json.dumps(
                    {
                        "body": "ABCD",
                        "query": {"q": "pandas"},
                        "headers": {"content-type": "application/json"},
                    }
                )
            )
        body, query, headers = hooks.get_request(local_file=local_file)
        assert body == "ABCD"
        assert query == {"q": "pandas"}
        assert headers == {"content-type": "application/json"}

    def test_send_response_with_local_file(self):
        local_file = "/tmp/response.json"
        hooks.send_response(
            body="ABCD",
            status_code=207,
            headers={"Session": 123},
            local_file=local_file,
        )

        with open(local_file) as f:
            content = f.read()
            response = json.loads(content)

        assert response["body"] == "ABCD"
        assert response["status_code"] == 207
        assert response["headers"] == {"Session": 123}

    def test_send_json_with_local_file(self):
        local_file = "/tmp/response.json"
        hooks.send_json(
            data={"ok": True},
            status_code=207,
            headers={"Session": 123},
            local_file=local_file,
        )

        with open(local_file) as f:
            content = f.read()
            response = json.loads(content)

        assert response["body"] == '{"ok": true}'
        assert response["status_code"] == 207
        assert response["headers"] == {
            "Session": 123,
            "content-type": "application/json",
        }



import traceback
from ...session import StaticSession


def run_hook(code, session: StaticSession):
    try:
        namespace = {}
        exec(code, namespace, namespace)
    except:
        traceback.print_exc()



from ..widget_base import Input


class KanbanBoardInput(Input):
    type = "kanban-board-input"

    def __init__(self, key: str, **kwargs):
        super().__init__(key)
        self.initial_value = kwargs.get("initial_value", {})
        self.label = kwargs.get("label", None)
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "initialValue": self.initial_value,
            "disabled": self.disabled,
        }

    def convert_answer(self, board: dict) -> str:
        return board



from abc import ABC, abstractmethod


class Resource(ABC):
    @staticmethod
    @abstractmethod
    def list(*args, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def add(*args, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def remove(*args, **kwargs):
        pass



from abstra_server.runtimes.dashes.program import PythonProgram
from abstra_server.api.classes import DashJSON, SlotJSON
import unittest

dash_page_state = {"widgets": {"widgetA": {"value": None}, "widgetB": {"value": None}}}


def prepare_program(slot_dict: dict):
    dash = DashJSON(
        path="foo",
        file=None,
        layout=dict(
            version="0.2",
            slot=slot_dict,
            props={},
        ),
        title="foo",
    )
    program = PythonProgram(dash, "print('hello world')")
    program.root = SlotJSON(slot_dict)
    return program


no_slots_layout = {
    "widgetA": {
        "id": "widgetA",
        "type": "text-input",
        "props": {"label": '"label"', "placeholder": '"placeholder"'},
        "events": {},
        "colEnd": 8,
        "rowEnd": 5,
        "colStart": 6,
        "rowStart": 3,
        "variable": None,
    }
}

one_grid_layout = {
    "gridBlock": {
        "id": "gridBlock",
        "type": "if-block",
        "row": 1,
        "height": 5,
        "order": 0,
        "props": {"condition": "True"},
        "slot": {
            "widgetA": {
                "type": "text-input",
                "props": {"label": '"label"', "placeholder": '"placeholder"'},
                "events": {},
                "rowStart": 2,
                "rowEnd": 4,
                "colStart": 6,
                "colEnd": 8,
                "id": "widgetA",
                "variable": None,
            }
        },
    },
}


class TestComputeWidgets(unittest.TestCase):
    def test_get_widget_simple_case(self):
        program = prepare_program(no_slots_layout)
        widget = program._PythonProgram__get_widget("widgetA")
        self.assertEqual(widget.__dict__, no_slots_layout["widgetA"])

    def test_get_widget_one_grid(self):
        program = prepare_program(one_grid_layout)
        widget = program._PythonProgram__get_widget("widgetA")
        self.assertEqual(
            widget.__dict__, one_grid_layout["gridBlock"]["slot"]["widgetA"]
        )



import os
import json

abstra_json = {"version": "0.2"}


def init_dir(path: str):
    abstra_json_path = os.path.join(path, "abstra.json")
    if not os.path.exists(path):
        os.makedirs(path)
    if not os.path.exists(abstra_json_path):
        with open(abstra_json_path, "w") as f:
            f.write(json.dumps(abstra_json))



import io
from typing import Union
from .apis import upload_file
from .types import PILImage
from tempfile import NamedTemporaryFile


def convert_file(file: Union[str, io.IOBase, PILImage]) -> str:
    if not file:
        return ""

    if isinstance(file, str):
        # URL or base64 encoded string
        if file.startswith("http") or file.startswith("data:"):
            return file

        # path to file
        file = open(file, "rb")

    if callable(getattr(file, "save", None)):
        file_path = "/tmp/img.png"
        file.save(file_path)
        file = open(file_path, "rb")

    return upload_file(file)


def download_file(url):
    import requests

    f = NamedTemporaryFile()
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        for chunk in r.iter_content(chunk_size=8192):
            f.write(chunk)
    f.seek(0)
    return f



from abstra.tables import get_db
import os
from uuid import uuid4 as uuid


class TestColumns:
    def test_update_not_null(self):
        # setup
        base_path = f"/tmp/{uuid()}"
        os.mkdir(base_path)
        os.environ["ABSTRA_DATABASE_URL"] = f"{base_path}/db.sqlite3"
        db = get_db()
        table = db.create_table()
        old_columns = db.get_columns(table.name)
        col = db.create_column(table.name)

        # experiment
        db.update_column(table.name, col.name, {"not_null": True})

        # result
        col2 = db.get_column(table.name, col.name)
        new_columns = db.get_columns(table.name)
        assert len(old_columns) + 1 == len(new_columns)
        assert col2.not_null == True



import unittest
from abstra_server.api import API
from uuid import uuid4 as uuid
from .fixtures import init_dir


class TestDashes(unittest.TestCase):
    def test_api_create(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        api = API(root=workspace_root_path)

        dashes_len = len(api.get_dashes())

        dash = api.create_dash()

        self.assertEqual(len(api.get_dashes()), dashes_len + 1)

        self.assertEqual(dash.layout.slot.__dict__, {})

    def test_api_update(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        api = API(root=workspace_root_path)

        dash = api.create_dash()

        new_layout = {
            "version": "0.2",
            "props": {},
            "slot": {
                "75838e71-92cd-4d8c-bfc0-f9ae23c32559": {
                    "events": {},
                    "id": "75838e71-92cd-4d8c-bfc0-f9ae23c32559",
                    "props": {"label": '"Click here!"'},
                    "position": {
                        "rowStart": 8,
                        "rowEnd": 9,
                        "colStart": 8,
                        "colEnd": 9,
                    },
                    "type": "click-input",
                    "variable": None,
                }
            },
        }

        updated_dash = api.update_dash(
            dash.path, dict(title="new title", layout=new_layout)
        )
        self.maxDiff = None
        self.assertEqual(updated_dash.title, "new title")
        self.assertEqual(updated_dash.layout.__dict__, new_layout)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
from dataclasses import dataclass
from typing import Union, Type, Tuple, Any, Optional
import traceback

from ...utils import convert_answer, revert_value
from .autocomplete import get_suggestions
from ...api.classes import DashJSON, SlotJSON, SlottableJSON, DashWidgetJSON
from abstra.widgets import (
    get_widget_class,
    is_prop_required,
    get_widget_name,
    get_prop_type,
    is_broker_prop_form_only,
    Input,
    Output,
)

WidgetClassType = Union[Type[Input], Type[Output]]


@dataclass()
class DashPageState(dict):
    timestamp: int
    widgets: dict


class PythonProgram:
    root: SlotJSON

    def __init__(self, dash: DashJSON, code: str) -> None:
        self.code = code
        self.root = dash.layout.slot

        # state: { [variable: string]: value }
        self.state = {}

    def ex(self, cmd: str):
        exec(cmd, self.state, self.state)

    def ev(self, expr: str):
        return eval(expr, self.state, self.state)

    def execute_initial_code(self):
        if not self.code:
            return

        self.ex(self.code)

    def set_variable(self, variable: str, value):
        try:
            self.state.update({"__temp_value__": value})
            self.ex(f"{variable} = __temp_value__")
        except Exception as e:
            traceback.print_exc()
        finally:
            self.state.pop("__temp_value__", None)

    def execute_widget_event(self, wid, type, payload, dash_page_state: DashPageState):
        widget = self.__get_widget(wid)
        cmd = widget.events.get(type)

        if not cmd:
            return False

        self.__execute_widget_event_command(widget, cmd, payload, dash_page_state)
        return True

    def __execute_widget_event_command(
        self, widget: DashWidgetJSON, cmd: str, payload, dash_page_state: DashPageState
    ):
        _, widget_value = self.__get_widget_context(widget, dash_page_state)

        self.state.update({"__widget__": widget_value})
        self.state.update({"__event__": {"value": widget_value, "payload": payload}})

        try:
            self.ex(cmd)
        except Exception as e:
            traceback.print_exc()
            return {"repr": traceback.format_exc()}
        finally:
            self.state.pop("__widget__", None)
            self.state.pop("__event__", None)

    def execute_widget_input(self, wid: str, dash_page_state: DashPageState):
        widget = self.__get_widget(wid)
        if not widget.variable:
            return False

        _, value = self.__get_widget_context(widget, dash_page_state)
        self.set_variable(widget.variable, value)
        return True

    def evaluate_widgets(self, dash_page_state: DashPageState):
        computed_widgets, computed_variables, errors = self.__compute_slot_widgets(
            self.root, dash_page_state
        )
        return {
            "props": computed_widgets,
            "variables": computed_variables,
            "errors": errors,
            "stateTimestamp": dash_page_state.get("timestamp"),
        }

    def __compute_slot_widgets(self, slot: SlotJSON, dash_page_state: DashPageState):
        computed_widgets = {}
        computed_variables = {}
        errors = {"props": {}, "variables": {}, "widgets": {}}

        for wid, widget in slot.items():
            if isinstance(widget, SlottableJSON):
                (
                    computed_slot,
                    computed_slot_variables,
                    errors_slot,
                ) = self.__compute_slot_widgets(widget.slot, dash_page_state)
                (
                    widget_props,
                    props_errors,
                    widget_errors,
                ) = self.__compute_slottable_props(widget)
                computed_widgets = {**computed_widgets, **computed_slot}
                computed_variables = {**computed_variables, **computed_slot_variables}
                computed_widgets[wid] = widget_props
                if widget_errors:
                    errors["widgets"][wid] = widget_errors
                if props_errors:
                    errors["props"][wid] = props_errors

                merge_errors(errors, errors_slot)
            elif isinstance(widget, DashWidgetJSON):
                variable, props, widget_errors = self.__compute_widget(
                    widget, dash_page_state
                )
                computed_widgets[wid] = props
                computed_variables[wid] = variable
                if widget_errors["props"]:
                    errors["props"][wid] = widget_errors["props"]
                if widget_errors["variables"]:
                    errors["variables"][wid] = widget_errors["variables"]
                if widget_errors["widgets"]:
                    errors["widgets"][wid] = widget_errors["widgets"]
            else:
                raise Exception(f"Invalid widget type {type(widget)}")

        return computed_widgets, computed_variables, errors

    def __compute_widget(self, widget: DashWidgetJSON, dash_page_state: DashPageState):
        widget_class, widget_value = self.__get_widget_context(widget, dash_page_state)
        self.state.update({"__widget__": widget_value})
        errors = {"props": {}, "widgets": {}, "variables": {}}

        variable, errors["variables"] = self.__compute_widget_variable(
            widget, widget_class
        )
        props, errors["props"], errors["widgets"] = self.__compute_widget_props(
            widget, widget_class
        )

        self.state.pop("__widget__", None)
        return variable, props, errors

    def __filter_form_only_props(self, widget: DashWidgetJSON, props: dict):
        return {
            k: v
            for k, v in props.items()
            if not is_broker_prop_form_only(widget.type, k)
        }

    def __compute_widget_props(
        self, widget: DashWidgetJSON, widget_class: WidgetClassType
    ):
        props = {"key": "key"}
        props_errors = {}
        widget_errors = None
        result = None
        for prop, expr in widget.props.items():
            if is_prop_required(widget.type, prop) and (
                expr.strip() == "" or expr is None
            ):
                prop_type = get_prop_type(widget.type, prop)
                widget_name = get_widget_name(widget.type)
                props_errors[prop] = {"repr": "Missing required prop"}
                widget_errors = {
                    "repr": f'Missing required prop "{prop}" ({prop_type}) for widget "{widget_name}".'
                }
                break

            try:
                props[prop] = self.ev(expr) if expr else None
            except Exception as e:
                traceback.print_exc()
                props_errors[prop] = {"repr": traceback.format_exc()}
        else:
            try:
                result = self.__filter_form_only_props(
                    widget, widget_class(**props).json()
                )
            except Exception as e:
                traceback.print_exc()
                widget_errors = {"repr": traceback.format_exc()}

        return result, props_errors, widget_errors

    def __compute_widget_variable(
        self, widget: DashWidgetJSON, widget_class: WidgetClassType
    ):
        if not widget.variable:
            return None, None

        try:
            # Check if it is a variable returning it's value
            self.ev(widget.variable)
            self.ex(f"{widget.variable} = {widget.variable}")
            variable_value = self.ev(widget.variable)
            return revert_value(widget_class, variable_value), None

        except Exception as e:
            traceback.print_exc()
            return None, {"repr": traceback.format_exc()}

    def __compute_slottable_props(self, widget: DashWidgetJSON):
        props = {}
        props_errors = {}
        widget_errors = None
        for prop, expr in widget.props.items():
            try:
                props[prop] = self.ev(expr) if expr else None
            except Exception as e:
                traceback.print_exc()
                props_errors[prop] = {"repr": traceback.format_exc()}

        return props, props_errors, widget_errors

    def __get_widget_context(
        self, widget: DashWidgetJSON, dash_page_state: DashPageState
    ) -> Tuple[WidgetClassType, Any]:
        cls = get_widget_class(widget.type)
        value = dash_page_state["widgets"].get(widget.id, {"value": None})["value"]
        converted_value = convert_answer(cls, value)
        return cls, converted_value

    def __get_widget_id_context(
        self, wid: str, dash_page_state: DashPageState
    ) -> Tuple[WidgetClassType, Any]:
        widget = self.__get_widget(wid)
        return self.__get_widget_context(widget, dash_page_state)

    def __get_widget(self, wid: str):
        widget = search_widget_in_slot(self.root, wid)
        if not widget:
            raise Exception(f"Widget {wid} not found")
        return widget

    def get_autocomplete_suggestions(self, added_code_snippet):
        return get_suggestions(self.code, added_code_snippet)


def search_widget_in_slot(slot: SlotJSON, wid: str) -> Optional[DashWidgetJSON]:
    if slot.get(wid):
        slottable = slot.get(wid)
        if isinstance(slottable, DashWidgetJSON):
            return slottable
        elif not isinstance(slottable, SlottableJSON):
            raise Exception(f"Invalid slottable type {type(slottable)}")

    for widget in slot.values():
        if not isinstance(widget, SlottableJSON):
            continue

        result = search_widget_in_slot(widget.slot, wid)
        if result:
            return result

    return None


def merge_errors(errors, errors_slot):
    errors["props"] = {**errors["props"], **errors_slot["props"]}
    errors["variables"] = {**errors["variables"], **errors_slot["variables"]}
    errors["widgets"] = {**errors["widgets"], **errors_slot["widgets"]}



from ...session import LiveSession
from .program import PythonProgram
from ...api.classes import DashJSON
from ...contract import dashes_contract
from .message_handler import MessageHandler
from simple_websocket import ConnectionError


def run_dash(session: LiveSession, dash: DashJSON, code: str):
    py = PythonProgram(dash, code)
    msg_handler = MessageHandler(py, session)
    session.send(dashes_contract.ExecutionIdMessage(session.id))
    while True:
        try:
            type, data = session.recv()
            msg_handler.handle(type, data)
        except ConnectionError:
            print("connection closed")



import inspect

from abstra.widgets import library
from .widget_base import Input, Output
from .metadata import metadata


output_types = [
    w for w in library.__dict__.values() if inspect.isclass(w) and issubclass(w, Output)
]

input_types = [
    w for w in library.__dict__.values() if inspect.isclass(w) and issubclass(w, Input)
]

widget_types = output_types + input_types


def get_widget_class(type: str):
    return [w for w in widget_types if w.type == type][0]


def _get_prop(widget_type: str, prop_name: str):
    params = metadata[widget_type]["pythonAPI"]["params"]
    return [p for p in params if p["argName"] == prop_name][0]


def is_prop_required(widget_type: str, prop_name: str):
    try:
        prop = _get_prop(widget_type, prop_name)
        return not prop["isKwarg"]
    except:
        return False


def _get_broker_prop(widget_type: str, prop_name: str):
    params = metadata[widget_type]["brokerAPI"]["params"]
    return [p for p in params if p["argName"] == prop_name][0]


def is_broker_prop_form_only(widget_type: str, prop_name: str):
    try:
        prop = _get_broker_prop(widget_type, prop_name)
        return prop["formOnly"]
    except:
        return False


def get_widget_name(widget_type: str):
    try:
        return metadata[widget_type]["name"]
    except:
        return None


def get_prop_type(widget_type: str, prop_name: str):
    try:
        prop = _get_prop(widget_type, prop_name)
        return prop["typeName"]
    except:
        return None



from abstra_cli.utils import remove_filepath_prefix


class TestFileUtils:
    def test_remove_filepath_prefix_no_dot(self):
        filepath = "folder/file"
        prefix = "folder"
        assert remove_filepath_prefix(filepath, prefix) == "file"

    def test_remove_filepath_prefix_dotted_both(self):
        filepath = "./folder/file"
        prefix = "./folder"
        assert remove_filepath_prefix(filepath, prefix) == "file"

    def test_remove_filepath_prefix_dotted_filepath(self):
        filepath = "./folder/file"
        prefix = "folder"
        assert remove_filepath_prefix(filepath, prefix) == "file"

    def test_remove_filepath_prefix_dotted_prefix(self):
        filepath = "folder/file"
        prefix = "./folder"
        assert remove_filepath_prefix(filepath, prefix) == "file"



import unittest
from abstra_server.api import API
from uuid import uuid4 as uuid
from .fixtures import init_dir


class TestWorkspace(unittest.TestCase):
    def test_api_update(self):
        workspace_root_path = f"/tmp/{uuid()}"

        init_dir(workspace_root_path)
        api = API(root=workspace_root_path)

        api.update_workspace({"name": "test-workspace-updated"})

        workspace = api.get_workspace()

        self.assertEqual(workspace.name, "test-workspace-updated")



import os
import shutil
from uuid import uuid4
from pathlib import PosixPath
from abstra_cli.utils import files_from_directory


def generate_random_folder():
    name = "dir-" + uuid4().hex
    path = "/tmp/" + name
    os.mkdir(path)
    return path


class TestListing:
    def add_file(self, name, content, path=None):
        filepath = (path or self.path) + "/" + name
        with open(filepath, "w", encoding="utf-8") as f:
            f.write(content)
        return filepath

    def add_folder(self, name, path=None):
        folderpath = (path or self.path) + "/" + name
        os.mkdir(folderpath)
        return folderpath

    def setup_method(self):
        self.path = generate_random_folder()
        os.chdir(self.path)

    def teardown_method(self):
        os.chdir("/tmp")
        shutil.rmtree(self.path)

    def test_empty_directory(self):
        files = files_from_directory(self.path)
        assert files == []

    def test_no_ignore(self):
        filepath = self.add_file("foo", "bar")
        files = files_from_directory(self.path)
        assert files == [PosixPath(filepath)]

    def test_ignore_file(self):
        self.add_file(".abstraignore", "ignored")
        self.add_file("ignored", "foo")
        tracked = self.add_file("tracked", "bar")
        files = files_from_directory(self.path)
        assert files == [PosixPath(tracked)]

    def test_ignore_empty(self):
        self.add_folder("ignored")
        self.add_folder("empty")
        self.add_file(".abstraignore", "ignored")
        files = files_from_directory(self.path)
        assert files == []

    def test_ignore_folder(self):
        ignored = self.add_folder("ignored")
        self.add_file("abc", "foo", path=ignored)
        ignored2 = self.add_folder("ignored2")
        self.add_file("xyz", "foo", path=ignored2)
        folder = self.add_folder("tracked")
        tracked = self.add_file("tracked", "tracked", path=folder)
        self.add_file(".abstraignore", "ignored\nignored2/")
        files = files_from_directory(self.path)
        assert files == [PosixPath(tracked)]

    def test_ignore_wildcard(self):
        self.add_file("ignored.ipynb", "foo")
        tracked = self.add_file("tracked", "bar")
        self.add_file(".abstraignore", "*.ipynb")
        files = files_from_directory(self.path)
        assert files == [PosixPath(tracked)]

    def test_ignore_abstra_folder(self):
        abstra = self.add_folder(".abstra")
        self.add_file("credentials", "SECRET", path=abstra)
        files = files_from_directory(self.path)
        assert files == []

    def test_ignore_git_folder(self):
        git = self.add_folder(".git")
        self.add_file("somefile", "somecontent", path=git)
        files = files_from_directory(self.path)
        assert files == []

    def test_ignore_abstra_ignore_file(self):
        self.add_file(".abstraignore", "ignored")
        files = files_from_directory(self.path)
        assert files == []

    def test_ignore_git_ignore_file(self):
        self.add_file(".gitignore", "ignored")
        files = files_from_directory(self.path)
        assert files == []



from ..apis import main as api_main
from . import utils as utils


def list_workspace_hooks():
    query = """
        query GetHooks {
            hooks {
                path
                title
                script {
                    enabled
                }
            }
        }
    """
    return api_main.hf_hasura_runner(query).get("hooks", [])


def add_workspace_hook(data):
    _, workspace_id, _ = api_main.get_auth_info()
    hook_data = {
        "title": data["name"],
        "script": {
            "data": {
                "workspace_id": workspace_id,
                "enabled": data.get("enabled", True),
                "code": data["code"],
                "name": data["name"],
            }
        },
    }

    data.pop("name", None)
    data.pop("code", None)
    data.pop("enabled", None)
    hook_data.update(data)

    query = """
        mutation InsertHook($hook_data: [hooks_insert_input!]!) {
            insert_hooks(
                objects: $hook_data
            ) {
                returning {
                    path
                    title
                }
            }
        }
    """

    return (
        api_main.hf_hasura_runner(query, {"hook_data": hook_data})
        .get("insert_hooks", {})
        .get("returning", {})[0]
    )


def update_workspace_hook(path, data):
    hook_data = data.copy()
    script_data = {}

    name = hook_data.pop("name", None)
    if name:
        hook_data["title"] = name
        script_data["name"] = name

    code = hook_data.pop("code", None)
    if code:
        script_data["code"] = code

    enabled = hook_data.pop("enabled", None)
    if enabled is not None:
        script_data["enabled"] = enabled

    request_data = {"path": path, "hook_data": hook_data, "script_data": script_data}
    update_query = """
        mutation UpdateHook($path: String!, $hook_data: hooks_set_input, $script_data: scripts_set_input = {}) {
            update_hooks(where: {path: {_eq: $path}}, _set: $hook_data) {
                returning {
                    id
                    path
                    title
                }
            }
            update_scripts(where: {hook: {path: {_eq: $path}}}, _set: $script_data) {
                returning {
                    name
                }
            }
        }
    """
    return api_main.hf_hasura_runner(update_query, request_data)


def upsert_workspace_hook(data):
    path = data["path"]

    query = """
        query FindHook($path: String!) {
            hooks(where: {path: {_eq: $path}}) {
                path
            }
        }
    """

    hooks = api_main.hf_hasura_runner(query, {"path": path}).get("hooks")
    if len(hooks):
        return update_workspace_hook(path, data)
    else:
        return add_workspace_hook(data)


def delete_workspace_hook(path):
    query = """
        mutation DeleteHook($path: String!) {
            delete_hooks(where: {path: {_eq: $path}}) {
                returning {
                    id
                    path
                    title
                }
            }
        }
    """

    return api_main.hf_hasura_runner(query, {"path": path})


def list_logs(limit, offset):
    query = """
        query GetHookLogs($limit: Int, $offset: Int) {
            hooks {
                logs(offset: $offset, limit: $limit, order_by: {created_at: desc}) {
                    id
                    hook_id
                    created_at
                    search_term
                    start_date
                    end_date
                    exit_code
                    executed_by
                    execution_id
                    request
                    response
                    trigger
                    status
                    execution {
                        status
                    }
                }
            }
        }
    """

    hooks = api_main.hf_hasura_runner(query).get("hooks", [])
    logs = utils.flat_items_logs(hooks, path_or_id="id")

    return {"logs": utils.sampling(logs, limit, offset)}


def list_logs_by_hook_id(id, limit, offset):
    query = """
        query GetHookLogs($limit: Int, $offset: Int, $id: uuid) {
            hooks(where: {id: {_eq: $id}}) {
                logs(offset: $offset, limit: $limit, order_by: {created_at: desc}) {
                    id
                    hook_id
                    created_at
                    search_term
                    start_date
                    end_date
                    exit_code
                    executed_by
                    execution_id
                    request
                    response
                    stderr_message
                    stdout_message
                    trigger
                    status
                    execution {
                        status
                    }
                }
            }
            }
    """
    hooks = api_main.hf_hasura_runner(
        query, {"id": id, "limit": limit, "offset": offset}
    ).get("hooks")
    if hooks:
        return {"logs": utils.flatten_list([hook.get("logs") for hook in hooks])}
    return {"logs": []}


def list_detailed_log_by_id(id):
    query = """
        query GetDetailedLog($id: uuid!) {
            hooks {
                logs(where: {id: {_eq: $id}}) {
                    stderr_message
                    stdout_message
                    status
                    execution {
                        id
                        status
                    }
                }
            }
        }
    """
    hooks = api_main.hf_hasura_runner(query, {"id": id}).get("hooks", None)
    if hooks:
        return {
            "detailed_log": utils.flatten_list([hook.get("logs") for hook in hooks])
        }



from collections import UserDict
from .debug_utils import CloseDTO, Frames


def get_connection():
    raise NotImplementedError("Must run inside server")


def send(data, frames: Frames = None):
    return get_connection().send(data, frames)


def receive(path: str = ""):
    return get_connection().receive(path)


def close(dto: CloseDTO):
    return get_connection().close(dto)


# this is a class to wrap all read methods of dicts to call get_url_params instead
class UrlParams(UserDict):
    @property
    def params(self):
        conn = get_connection()
        return conn.url_params if conn else {}

    def __getitem__(self, key):
        return self.params.__getitem__(key)

    def __iter__(self):
        return self.params.__iter__()

    def __len__(self) -> int:
        return self.params.__len__()

    def __contains__(self, key):
        return self.params.__contains__(key)

    def __str__(self) -> str:
        return self.params.__str__()

    def __repr__(self) -> str:
        return self.params.__repr__()

    def keys(self):
        return self.params.keys()

    def values(self):
        return self.params.values()

    def items(self):
        return self.params.items()

    def has_key(self, key):
        return key in self.params


url_params = UrlParams()



import base64, simplejson, jwt, os
from traceback import StackSummary


def serialize(obj):
    return simplejson.dumps(obj, ignore_nan=True)


def deserialize(st):
    return simplejson.loads(st)


def is_serializable(st):
    try:
        serialize(st)
        return True
    except:
        return False


def btos(b64):
    return base64.b64decode(b64).decode()


def get_staticmethod(cls, name):
    # hack because python sucks
    method_key = f"_{cls.__name__}{name}"
    return getattr(cls, method_key, None)


def convert_answer(cls, value):
    convert = get_staticmethod(cls, "__convert_answer")
    return convert(value) if convert else value


def revert_value(cls, value):
    revert = get_staticmethod(cls, "__revert_value")
    return revert(value) if revert else value


def read_file(filename: str):
    with open(filename, "r", encoding="utf8") as f:
        return f.read()


def filter_object(object, filter_func):
    filtered = {}
    for key, value in object.items():
        if filter_func(value):
            filtered[key] = value
    return filtered


def prepate_traceback(tb: StackSummary):
    for i, stack in enumerate(tb):
        if stack.filename == "<string>":
            return [{"line": x.lineno, "filename": x.filename} for x in tb[i:]]
    else:
        return []


def decode_jwt(jwt_str: str):
    try:
        if os.getenv("PASSWORDLESS_JWT_SECRET"):
            return jwt.decode(
                jwt_str, key=os.getenv("PASSWORDLESS_JWT_SECRET"), algorithms=["HS256"]
            )
        return jwt.decode(
            jwt_str, options={"verify_signature": False}, algorithms=["HS256"]
        )
    except:
        return None


def random_id(length=10):
    import random, string

    return "".join(random.choices(string.ascii_letters + string.digits, k=length))



import os, sys
from .utils import ABSTRA_FOLDER, CREDENTIALS_FILE

overwriten_auth_headers = {}


def overwrite_auth_headers(headers):
    if not isinstance(headers, dict):
        raise ValueError("headers must be a dict")
    global overwriten_auth_headers
    overwriten_auth_headers = headers


def create_abstra_dir():
    if not os.path.exists(ABSTRA_FOLDER):
        os.makedirs(ABSTRA_FOLDER)


def save_credentials(api_token: str):
    create_abstra_dir()
    with open(CREDENTIALS_FILE, "w") as f:
        f.write(api_token.strip())


def get_credentials():
    if os.getenv("ABSTRA_API_TOKEN"):
        return os.getenv("ABSTRA_API_TOKEN")

    if not os.path.exists(CREDENTIALS_FILE):
        return None

    with open(CREDENTIALS_FILE) as f:
        return f.read().strip()


def get_auth_headers(api_token=None):
    if overwriten_auth_headers:
        return {"content-type": "application/json", **overwriten_auth_headers}

    api_token = api_token or get_credentials()
    if api_token:
        return {"content-type": "application/json", "API-Authorization": api_token}

    print("No API token found. Please login with `abstra login`")
    return None



from typing import Any, Optional, List
import sqlite3
from dataclasses import dataclass
import re


@dataclass
class Column:
    id: int
    name: str
    type: str
    not_null: bool
    primary_key: bool
    default: Optional[str] = None

    @property
    def editor_dto(self):
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "not_null": self.not_null,
            "default": self.default,
            "primary_key": self.primary_key,
        }


@dataclass
class Table:
    name: str
    columns: List[Column]

    @property
    def editor_dto(self):
        return {"name": self.name, "columns": [c.editor_dto for c in self.columns]}


def random_id(length=10):
    import random, string

    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


def get_columns_from_table(cur: sqlite3.Cursor, table_name: str):
    rows = cur.execute(
        f"""
        pragma table_info("{table_name}")
    """
    ).fetchall()

    return [
        Column(
            id=id,
            name=name,
            type=type,
            not_null=not_null == 1,
            default=default,
            primary_key=primary_key == 1,
        )
        for (id, name, type, not_null, default, primary_key) in rows
    ]


def get_column(cur: sqlite3.Cursor, table_name: str, column_name: str):
    return [
        c for c in get_columns_from_table(cur, table_name) if c.name == column_name
    ][0]


def transform_expression(exp: str, dict_params: dict[str, Any]):
    param_key = r":\w+"
    params = []
    if exp:
        for match in re.findall(param_key, exp):
            params.append(dict_params[match[1:]])
            exp = exp.replace(match, "?")
        return f"WHERE {exp}", params
    else:
        return "", params


class SqliteDB:
    def __init__(self, db_path: str) -> None:
        self.db_path = db_path

    def connect(self, results="dicts"):
        conn = sqlite3.connect(self.db_path)
        if results == "dicts":
            conn.row_factory = dict_factory
        return conn

    def create_table(self):
        cur = self.connect("tuples").cursor()

        table_name = f"table-{random_id()}"
        cur.execute(
            f"""
            CREATE TABLE "{table_name}" (
                id INTEGER PRIMARY KEY NOT NULL,
                created_at INTEGER DEFAULT CURRENT_TIMESTAMP,
                updated_at INTEGER DEFAULT CURRENT_TIMESTAMP
            );
        """
        )

        return Table(
            name=table_name,
            columns=[
                Column(
                    id=0,
                    name="id",
                    type="INTEGER",
                    not_null=True,
                    default=None,
                    primary_key=True,
                ),
                Column(
                    id=1,
                    name="created_at",
                    type="INTEGER",
                    not_null=True,
                    default="CURRENT_TIMESTAMP",
                    primary_key=False,
                ),
                Column(
                    id=2,
                    name="updated_at",
                    type="INTEGER",
                    not_null=True,
                    default="CURRENT_TIMESTAMP",
                    primary_key=False,
                ),
            ],
        )

    def update_table(self, name: str, changes: dict[str, str]):
        cur = self.connect("tuples").cursor()

        if "name" in changes:
            old_name = name
            new_name = changes["name"]

            cur.execute(
                f"""
                ALTER TABLE "{old_name}"
                RENAME TO "{new_name}"
            """
            )

            return Table(name=new_name, columns=get_columns_from_table(cur, name))

    def get_table(self, name: str):
        cur = self.connect("tuples").cursor()

        return Table(name=name, columns=get_columns_from_table(cur, name))

    def get_tables(self):
        cur = self.connect("tuples").cursor()
        rows = cur.execute(
            """
            SELECT name
            FROM sqlite_schema
            WHERE type='table'
            ORDER BY name
        """
        ).fetchall()

        return [
            Table(name=name, columns=get_columns_from_table(cur, name))
            for (name,) in rows
        ]

    def delete_table(self, name: str):
        cur = self.connect("tuples").cursor()
        cur.execute(
            f"""
            DROP TABLE "{name}"
        """
        )

    def duplicate_table(self, name: str):
        cur = self.connect("tuples").cursor()
        new_name = f"{name}-{random_id()}"
        cur.execute(
            f"""
            CREATE TABLE "{new_name}" AS
            SELECT * FROM "{name}"
        """
        )
        return Table(name=new_name, columns=get_columns_from_table(cur, new_name))

    def db_types(self):
        return [
            {"type": "NULL", "input_spec": {"type": "hidden"}},
            {"type": "INTEGER", "input_spec": {"type": "number"}},
            {"type": "REAL", "input_spec": {"type": "number"}},
            {"type": "INT", "input_spec": {"type": "number"}},
            {"type": "TEXT", "input_spec": {"type": "text"}},
            {"type": "BLOB", "input_spec": {"type": "file"}},
        ]

    def get_column(self, table_name: str, column_name: str) -> Optional[Column]:
        cur = self.connect("tuples").cursor()
        return get_column(cur, table_name, column_name)

    def get_columns(self, table_name: str):
        cur = self.connect("tuples").cursor()
        rows = cur.execute(
            f"""
            pragma table_info("{table_name}")
        """
        ).fetchall()

        return [
            Column(
                id=id,
                name=name,
                type=type,
                not_null=not_null == 1,
                default=default,
                primary_key=primary_key == 1,
            )
            for (id, name, type, not_null, default, primary_key) in rows
        ]

    def update_column(self, table_name: str, column_name: str, changes: dict[str, str]):
        cur = self.connect("tuples").cursor()

        column = get_column(cur, table_name, column_name)

        if ("type" in changes) or ("not_null" in changes) or ("default" in changes):
            new_column_name = changes.get("name", column_name)
            new_column_type = changes.get("type", column.type)
            new_column_not_null = changes.get("not_null", column.not_null)
            new_column_default = changes.get("default", column.default)

            if new_column_not_null:
                not_null_exp = "NOT NULL"
            else:
                not_null_exp = ""

            if new_column_default:
                default_exp = f"DEFAULT {new_column_default}"
            else:
                default_exp = ""

            uuid = random_id()
            temp_column_name = f"{new_column_name}_{uuid}"

            cur.executescript(
                f"""
                ALTER TABLE "{table_name}"
                ADD COLUMN "{temp_column_name}" "{new_column_type}"
                {default_exp}
                {not_null_exp};

                UPDATE "{table_name}"
                SET "{temp_column_name}" = "{column_name}";
                
                ALTER TABLE "{table_name}"
                DROP COLUMN "{column_name}";
                
                ALTER TABLE "{table_name}"
                RENAME COLUMN "{temp_column_name}" TO "{new_column_name}"
            """
            )
        elif "name" in changes:
            new_column_name = changes["name"]
            new_column_type = column.type
            cur.execute(
                f"""
                ALTER TABLE "{table_name}"
                RENAME COLUMN "{column_name}" TO "{new_column_name}"
            """
            )
        else:
            raise Exception("Nothing changed")

        return Column(
            id=column.id,
            name=changes.get("name", column.name),
            type=changes.get("type", column.type),
            not_null=changes.get("not_null", column.not_null),
            default=changes.get("default", column.default),
            primary_key=changes.get("primary_key", column.primary_key),
        )

    def delete_column(self, table_name: str, column_name: str):
        cur = self.connect("tuples").cursor()

        cur.execute(
            f"""
            ALTER TABLE "{table_name}"
            DROP COLUMN "{column_name}"
        """
        )

    def create_column(self, table_name: str):
        cur = self.connect("tuples").cursor()

        column_name = f"column_{random_id()}"
        column_type = "TEXT"

        cur.execute(
            f"""
            ALTER TABLE "{table_name}"
            ADD COLUMN "{column_name}" "{column_type}"
        """
        )

        return get_column(cur, table_name, column_name)

    def select(
        self,
        table_name: str,
        where: str = "true",
        rows: list[str] = ["*"],
        **kwargs,
    ) -> list[dict[str, Any]]:
        with self.connect() as conn:
            where_exp, params = transform_expression(where, kwargs)
            rows_exp = ", ".join(rows)
            query = f'SELECT {rows_exp} FROM "{table_name}" {where_exp}'
            print(query)
            return conn.execute(query, params).fetchall()

    def insert(self, table_name: str, values: dict[str, Any] = {}) -> dict[str, Any]:
        with self.connect() as conn:
            if len(values) == 0:
                query = f'INSERT INTO "{table_name}" DEFAULT VALUES RETURNING *'
                result = conn.execute(query)
            else:
                params = [value for key, value in values.items()]
                keys = ", ".join([key for key, value in values.items()])
                values_exp = ", ".join(["?" for key, value in values.items()])
                query = f'INSERT INTO "{table_name}" ({keys}) VALUES ({values_exp}) RETURNING *'
                result = conn.execute(query, params).fetchone()

            return result

    def update(
        self, table_name: str, where: str, set: dict[str, Any] = {}, **kwargs
    ) -> None:
        with self.connect() as conn:
            params = [value for key, value in set.items()]
            set_exp = ", ".join([f"{key} = ?" for key, value in set.items()])
            where_exp, where_params = transform_expression(where, kwargs)
            params.extend(where_params)
            query = f'UPDATE "{table_name}" SET {set_exp} {where_exp} RETURNING *'
            result = conn.execute(query, params).fetchone()

            return result

    def delete(self, table_name: str, where: str, **kwargs) -> None:
        with self.connect() as conn:
            where_exp, params = transform_expression(where, kwargs)
            query = f'DELETE FROM "{table_name}" {where_exp} RETURNING *'
            result = conn.execute(query, params).fetchone()

            return dict(result=result)



import json, sys
from .resources import (
    Forms,
    Files,
    Packages,
    Vars,
    Hooks,
    Jobs,
    Dashes,
    Workspaces,
)

from . import messages


ACCEPTED_KEYS = ["files", "workspace", "forms", "hooks", "jobs", "packages", "vars"]


def get_abstra_json_path(parameters: dict) -> str:
    return parameters.get("file") or parameters.get("f") or "abstra.json"


def evaluate_parameters_file(parameters: dict) -> dict:
    file = get_abstra_json_path(parameters)

    data = None
    try:
        with open(file) as f:
            data = json.loads(f.read())
            if not isinstance(data, dict):
                print("Bad data")
                sys.exit(1)
    except:
        print("Deploy file not found or not correct format")
        sys.exit(1)

    for key in data.keys():
        if key not in ACCEPTED_KEYS:
            print(f"Extra data in deploy file {file} not accepted")
            sys.exit(1)

    return data, file


def deploy(**kwargs):
    abstra_json_data, abstra_json_path = evaluate_parameters_file(kwargs)

    dashes = Dashes.map_deploy_data(abstra_json_path, abstra_json_data)
    if len(dashes):
        messages.start_dashes_deploy()
        for dash_props in dashes:
            Dashes.add(upsert=True, **dash_props)

        remote_dashes_files = Files.list_dashes_files(".")
        local_dashes_files = [
            d["code"].replace(".py", ".abstradash.json") for d in dashes
        ]
        deleted_dashes_files = [
            f for f in remote_dashes_files if f not in local_dashes_files
        ]

        for deleted_dash_file in deleted_dashes_files:
            deleted_dash_path = deleted_dash_file.replace(
                ".abstradash.json", ""
            ).replace(f"./", "")
            Dashes.remove(deleted_dash_path)
            Files.remove(
                deleted_dash_file, deleted_dash_file.replace(".abstradash.json", ".py")
            )
    else:
        messages.no_dashes_to_deploy()
        pass

    forms = abstra_json_data.pop("forms", None)
    if forms:
        messages.start_forms_deploy()
        for form in forms:
            Forms.add(upsert=True, **form)
    else:
        messages.no_forms_to_deploy()
        pass

    hooks = abstra_json_data.pop("hooks", None)
    if hooks:
        messages.start_hooks_deploy()
        for hook in hooks:
            Hooks.add(upsert=True, **hook)
    else:
        messages.no_hooks_to_deploy()
        pass

    jobs = abstra_json_data.pop("jobs", None)
    if jobs:
        messages.start_jobs_deploy()
        for job in jobs:
            Jobs.add(upsert=True, **job)
    else:
        messages.no_jobs_to_deploy()
        pass

    files = abstra_json_data.pop("files", ["."])  # review security implications
    if files:
        messages.start_files_deploy()
        if isinstance(files, dict):
            Files.add(**files)
        elif isinstance(files, list):
            Files.add(*files)
    else:
        messages.no_files_to_deploy()
        pass

    packages = abstra_json_data.pop("packages", {"requirement": "requirements.txt"})
    if packages:
        messages.start_packages_deploy()
        if isinstance(packages, dict):
            Packages.add(**packages)
        elif isinstance(packages, list):
            Packages.add(*packages)
    else:
        messages.no_packages_to_deploy()
        pass

    vars = abstra_json_data.pop("vars", None)
    if vars:
        messages.start_vars_deploy()
        if isinstance(vars, dict):
            Vars.add(**vars)
        elif isinstance(vars, list):
            Vars.add(*vars)
    else:
        messages.no_vars_to_deploy()
        pass

    workspace_deploy_data = Workspaces.map_deploy_data(abstra_json_data)
    if workspace_deploy_data:
        Workspaces.update(**workspace_deploy_data)



import traceback
from ...contract import dashes_contract as contract
from ...api.classes import SlotJSON
from ...session import LiveSession
from .program import PythonProgram, DashPageState


def preview_only_check(method):
    def wrapper(self: "MessageHandler", *args, **kwargs):
        if self.session.is_preview:
            method(self, *args, **kwargs)
            return

        self.session.send(
            contract.ErrorMessage("This method is only allowed in preview mode")
        )

    return wrapper


class MessageHandler:
    py: PythonProgram
    session: LiveSession
    dash_page_state: DashPageState

    def __init__(self, py: PythonProgram, session: LiveSession) -> None:
        self.py = py
        self.session = session
        self.dash_page_state = None

    def handle(self, type: str, data):
        handlers = {
            "client-start": self.start,
            "widget-input": self.widget_input,
            "widget-event": self.widget_event,
            "browser:try-disconnect": self.log_only,
            "eval": self.eval,
            "widgets-changed": self.widgets_changed,
            "variable-created": self.variable_created,
            "autocomplete:load": self.autocomplete_load,
        }
        handler = handlers.get(type, self.default_handler)
        self.dash_page_state = data.get("state", self.dash_page_state)
        handler(data)

    def default_handler(self, _data):
        self.session.send(contract.ErrorMessage("unknown type"))

    def start(self, data):
        # data: { type: client-start, state: PAGESTATE, params: PARAMS }
        self.session.context["query_params"] = data.get("params", {})
        try:
            self.py.execute_initial_code()
        except Exception as e:
            tb = traceback.extract_tb(e.__traceback__)
            self.session.send(
                contract.ProgramStartFailedMessage(traceback.format_exc(), tb)
            )
            self.session.close()

        self._compute_and_send_widgets_props()

    def widget_input(self, data):
        # data: { type: widget-input, widgetId: string, state: PAGESTATE }
        if self.py.execute_widget_input(data["widgetId"], self.dash_page_state):
            self._compute_and_send_widgets_props()

    def widget_event(self, data):
        # data: { type: widget-event, widgetId: string, event: { type: string, payload: any }, state: PAGESTATE }
        if self.py.execute_widget_event(
            data["widgetId"],
            data["event"]["type"],
            data["event"].get("payload", {}),
            self.dash_page_state,
        ):
            self._compute_and_send_widgets_props()

    def log_only(self, _data):
        # data: { type: 'browser:try-disconnect', ... }
        # print("browser:try-disconnect", data)
        pass

    @preview_only_check
    def eval(self, data):
        # data: {type: eval, expression: string}
        try:
            try:
                value = self.py.ev(data["expression"])
                self.session.send(contract.EvalReturnMessage(repr(value)))
            except SyntaxError:
                self.py.ex(data["expression"])
                self.session.send(contract.EvalReturnMessage(""))
        except Exception as e:
            self.session.send(contract.EvalErrorMessage(traceback.format_exc()))

        self._compute_and_send_widgets_props()

    @preview_only_check
    def widgets_changed(self, data):
        # data: { type: widgets-changed, dashDefinition: { slot }, state }
        self.py.root = SlotJSON(data["dashDefinition"]["slot"])
        self._compute_and_send_widgets_props()

    @preview_only_check
    def variable_created(self, data):
        # data: { type: variable-created, name, value?, state }
        self.py.set_variable(data["name"], data.get("value"))
        self._compute_and_send_widgets_props()

    @preview_only_check
    def autocomplete_load(self, data):
        # data: { type: autocomplete:load, suggestionsFor: string, code: string }
        try:
            suggestions = self.py.get_autocomplete_suggestions(data["code"])
        except Exception as e:
            suggestions = []

        self.session.send(
            contract.AutocompleteSuggestionsMessage(
                suggestions=suggestions,
                suggestionsFor=data["suggestionsFor"],
            )
        )

    def _compute_and_send_widgets_props(self):
        try:
            computed = self.py.evaluate_widgets(self.dash_page_state)
            self.session.send(contract.WidgetsComputedMessage(**computed))
        except Exception as e:
            self.session.send(
                contract.WidgetsComputedGeneralErrorMessage(traceback.format_exc())
            )



# This is an interface only file
# Actual functions are defined on runtime


class AuthResponse:
    email: str


def get_user() -> AuthResponse:
    pass


def redirect(url: str, params={}):
    pass


def get_query_params() -> dict:
    pass


def alert(message: str, severity: str = "info"):
    pass


def execute_js():
    pass



import requests
from .utils import get_headers

BASE_URL = "https://tables.abstra.cloud"


class Tables:
    def __init__(self, api_key=None):
        self.api_key = api_key

    def statement(self, id):
        return Statements(id, self.api_key)

    def run_statement(self, id, params=None):
        statement = self.statement(id)
        return statement.run(params)


class Statements:
    def __init__(self, id, api_key=None):
        self.id = id
        self.api_key = api_key

    def run(self, params=None):
        if params is None:
            params = {}
        res = requests.post(
            f"{BASE_URL}/execute/{self.id}",
            json=params,
            headers=get_headers(self.api_key),
        )
        if res.status_code == 200:
            return res.json()
        else:
            raise Exception(res.text)



import flask, flask_sock

from ..api import API
from ..session import LiveSession, StaticSession
from ..runtimes import runners, run_hook
from .utils import send_from_dist


def get_player_bp(api: API):
    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)

    @bp.route("/_api/<path:id_or_path>", methods=["GET"])
    def get_runner_data(id_or_path):
        runtime = api.get_page_runtime(id_or_path)
        if not runtime:
            print("404", id_or_path)
            flask.abort(404)

        workspace = api.get_workspace()
        res = {}
        res[runtime.runner_type] = {
            **runtime.runner_dto,
            "workspace": workspace.runner_dto,  # TODO: fix this in frontend
        }
        return res

    @sock.route("/_socket")
    def websocket(conn: flask_sock.Server):
        is_preview = flask.request.args.get("isPreview")
        dash_path = flask.request.args.get("dashPath")
        form_path = flask.request.args.get("formPath")
        path = dash_path if dash_path is not None else form_path
        runtime = api.get_page_runtime(path)

        if not runtime:
            conn.close(message="Not found")
            return

        run = runners[runtime.runner_type]
        if runtime.file:
            code = api.read_text_file(runtime.file)  # TODO: handle 404
        else:
            code = ""
        session = LiveSession(conn, is_preview)

        try:
            run(session, runtime, code)
        finally:
            session.end()

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook(path):
        runtime = api.get_page_runtime(path)
        if not runtime:
            flask.abort(404)

        if not runtime.file:
            flask.abort(500)

        code = api.read_text_file(runtime.file)  # TODO: handle 404
        session = StaticSession()
        session.context["request"] = (
            flask.request.get_data(as_text=True),
            flask.request.args,
            flask.request.headers,
        )

        run_hook(code, session)

        body, status, headers = session.context.get("response", ({}, 200, {}))
        return flask.Response(status=status, headers=headers, response=body)

    @bp.route("/<path:filename>", methods=["GET"])
    def spa(filename: str):
        return send_from_dist(filename, "player.html")

    return bp



import unittest
from abstra_server.api import API
from uuid import uuid4 as uuid
from .fixtures import init_dir


class TestHooks(unittest.TestCase):
    def test_api_list(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        api = API(root=workspace_root_path)

        self.assertEqual(len(api.get_hooks()), 0)

        api.create_hook()

        self.assertEqual(len(api.get_hooks()), 1)


if __name__ == "__main__":
    unittest.main()



from .metadata import metadata


def types_compatible(prop, schema):
    if type(schema["typeName"]) == list:
        return any(
            types_compatible(prop, {**schema, "typeName": type})
            for type in schema["typeName"]
        )

    if schema["typeName"] == "string":
        return type(prop) == str
    if schema["typeName"] == "number":
        return type(prop) == int or type(prop) == float
    if schema["typeName"] == "boolean":
        return type(prop) == bool
    if schema["typeName"] == "array":
        return type(prop) == list
    if schema["typeName"] == "object":
        return type(prop) == dict
    if schema["typeName"] == "null":
        return prop == None
    if schema["typeName"] == "any":
        return True
    raise Exception(f"Unknown type: {schema['typeName']}")


def valid_prop(prop, schema):
    valid = True

    if schema.get("oneOf"):
        valid = valid and prop in schema.get("oneOf")

    return valid


def to_generic_type(type):
    if type == str:
        return "string"
    if type == int:
        return "number"
    if type == bool:
        return "boolean"
    if type == list:
        return "array"
    if type == dict:
        return "object"
    if type == None:
        return "null"
    raise Exception(f"Unknown type: {type}")


def validate_widget_props(widget):
    metadata_widget = metadata[widget["type"]]
    required_props = [
        prop
        for prop in metadata_widget["brokerAPI"]["params"]
        if not prop.get("isOptional", False)
    ]

    for prop in required_props:
        prop_name = prop["argName"]
        if not prop_name in widget:
            raise Exception(
                f"Error in {widget['type']}: {prop_name} not in {widget.keys()}"
            )
        assert types_compatible(widget[prop_name], prop)
        assert valid_prop(widget[prop_name], prop)

    optional_props = [
        prop
        for prop in metadata_widget["brokerAPI"]["params"]
        if prop.get("isOptional", False)
    ]
    for prop in optional_props:
        prop_name = prop["argName"]
        if not (
            widget.get(prop_name) is None
            or (
                types_compatible(widget[prop_name], prop)
                and valid_prop(widget[prop_name], prop)
            )
        ):
            print(types_compatible(widget[prop_name], prop))
            print(widget.get(prop_name))
            print(valid_prop(widget[prop_name], prop))
            raise Exception(
                f"{prop_name}: {type(widget[prop_name])} is not compatible with {prop}"
            )
    for prop in widget:
        if prop == "type":
            continue
        if not (
            prop in param["argName"] for param in metadata_widget["brokerAPI"]["params"]
        ):
            raise Exception(
                f'{prop} not in {metadata_widget["brokerAPI"]["params"].keys()}'
            )



from .. import credentials as credentials
from ..apis import main as api_main
import requests

billing_url = "https://billing.abstra.cloud"


def get_account_info():
    headers = credentials.get_auth_headers()
    _, workspace_id, _ = api_main.get_auth_info()
    response = requests.get(
        f"{billing_url}/workspaces/{workspace_id}/account", headers=headers
    )
    response_json = response.json()
    if response_json is None:
        return None
    return response_json



import json, pathlib, os, typing
import webbrowser
from .classes import AbstraJSON, DashJSON, FormJSON, WorkspaceJSON, HookJSON, JobJSON
from ..utils import random_id
from abstra.tables import get_db


class API:
    def __init__(self, root: str):
        cwd = pathlib.Path.cwd()
        self.root_path = (cwd / root).resolve().as_posix()
        self.abstra_json_path = (cwd / root / "abstra.json").resolve().as_posix()
        db_path = os.getenv(
            "ABSTRA_DATABASE_URL", os.path.join(self.root_path, "db.sqlite3")
        )
        os.environ["ABSTRA_DATABASE_URL"] = db_path
        self.db = get_db()

        try:
            self.__get_abstra_json()
        except FileNotFoundError:
            self.init_empty()

    def init_empty(self):
        self.persist(AbstraJSON.make_empty())

    def persist(self, abstra_json: AbstraJSON):
        with open(self.abstra_json_path, "w") as f:
            json.dump(abstra_json.__dict__, f, indent=2)

    def __get_abstra_json(self) -> AbstraJSON:
        with open(self.abstra_json_path, "r") as f:
            abstra_json_content = json.load(f)
        return AbstraJSON.from_dict(abstra_json_content)

    def read_text_file(self, path) -> str:
        with open(os.path.join(self.root_path, path), "r") as f:
            return f.read()

    def get_workspace(self) -> WorkspaceJSON:
        abstra_json = self.__get_abstra_json()
        return abstra_json.workspace

    def get_page_runtime(self, path) -> typing.Union[DashJSON, FormJSON, None]:
        abstra_json = self.__get_abstra_json()

        for form in abstra_json.forms:
            if path == form.path:
                return form

        for dash in abstra_json.dashes:
            if path == dash.path:
                return dash

        for hook in abstra_json.hooks:
            if path == hook.path:
                return hook

        return None

    # Forms CLRUD

    def create_form(self) -> FormJSON:
        abstra_json = self.__get_abstra_json()

        file_name = f"new_form_{random_id()}"
        file_path = f"{file_name}.py"

        with open(os.path.join(self.root_path, file_path), "w") as f:
            f.write(
                """from abstra.forms import display, read


name = read("What is your name?")
display(f"Hello World, {name}")"""
            )

        form = FormJSON(file=file_path, path=file_name, title="Untitled Form")

        abstra_json.forms.append(form)

        self.persist(abstra_json)

        return form

    def get_forms(self) -> list[FormJSON]:
        abstra_json = self.__get_abstra_json()
        return abstra_json.forms

    def get_form(self, path: str) -> typing.Optional[FormJSON]:
        abstra_json = self.__get_abstra_json()

        for form in abstra_json.forms:
            if path == form.path:
                return form

        return None

    def update_form(
        self, path: str, changes: dict[str, typing.Any]
    ) -> typing.Optional[FormJSON]:
        abstra_json = self.__get_abstra_json()
        forms = [f for f in abstra_json.forms if f.path == path]
        if len(forms) == 0:
            raise Exception("Form not found for path ", path)
        form = forms[0]
        form.update(changes)

        self.persist(abstra_json)

        return form

    def delete_form(self, path: str):
        abstra_json = self.__get_abstra_json()
        forms = abstra_json.forms
        abstra_json.forms = [f for f in forms if f.path != path]

        self.persist(abstra_json)

    # Dashes CLRUD

    def create_dash(self) -> DashJSON:
        abstra_json = self.__get_abstra_json()

        file_name = f"new_dash_{random_id()}"
        file_path = f"{file_name}.py"

        with open(os.path.join(self.root_path, file_path), "w") as f:
            f.write("foo = 'bar'")

        dash = DashJSON(
            file=file_path,
            path=f"new-dash_{random_id()}",
            title="Untitled Dash",
            layout=dict(version="0.2", props={}, slot={}),
        )

        abstra_json.dashes.append(dash)

        self.persist(abstra_json)

        return dash

    def get_dashes(self) -> list[DashJSON]:
        abstra_json = self.__get_abstra_json()
        return abstra_json.dashes

    def get_dash(self, path: str) -> typing.Optional[DashJSON]:
        abstra_json = self.__get_abstra_json()

        for dash in abstra_json.dashes:
            if path == dash.path:
                return dash

        return None

    def update_dash(
        self, path: str, changes: dict[str, typing.Any]
    ) -> typing.Optional[DashJSON]:
        abstra_json = self.__get_abstra_json()
        dashes = [d for d in abstra_json.dashes if d.path == path]
        if len(dashes) == 0:
            raise Exception("Dash not found for path ", path)
        dash = dashes[0]
        dash.update(changes)

        self.persist(abstra_json)

        return dash

    def delete_dash(self, path: str):
        abstra_json = self.__get_abstra_json()
        dashes = abstra_json.dashes
        abstra_json.dashes = [d for d in dashes if d.path != path]

        self.persist(abstra_json)

    # Hooks CLRUD

    def create_hook(self) -> HookJSON:
        abstract_json = self.__get_abstra_json()

        file_name = f"new_hook_{random_id()}"
        file_path = f"{file_name}.py"

        with open(os.path.join(self.root_path, file_path), "w") as f:
            f.write("print('Hello World')")

        hook = HookJSON(file=file_path, path=file_name, title="Untitled Hook")

        abstract_json.hooks.append(hook)

        self.persist(abstract_json)

        return hook

    def get_hooks(self) -> list[HookJSON]:
        abstra_json = self.__get_abstra_json()
        return abstra_json.hooks

    def get_hook(self, path: str) -> typing.Optional[HookJSON]:
        abstra_json = self.__get_abstra_json()

        for hook in abstra_json.hooks:
            if path == hook.path:
                return hook

        return None

    def update_hook(
        self, path: str, changes: dict[str, typing.Any]
    ) -> typing.Optional[HookJSON]:
        abstra_json = self.__get_abstra_json()
        hooks = [h for h in abstra_json.hooks if h.path == path]
        if len(hooks) == 0:
            raise Exception("Hook not found for path ", path)
        hook = hooks[0]
        hook.update(changes)
        self.persist(abstra_json)

        return hook

    def delete_hook(self, path: str):
        abstra_json = self.__get_abstra_json()
        hooks = abstra_json.hooks
        abstra_json.hooks = [h for h in hooks if h.path != path]

        self.persist(abstra_json)

    def open_file(self, file_path: str, create_if_not_exists: bool = False):
        complete_file_path = os.path.join(self.root_path, file_path)

        if (
            create_if_not_exists
            and file_path.endswith(".py")
            and not os.path.isfile(complete_file_path)
        ):
            with open(complete_file_path, "w") as f:
                f.write("")

        webbrowser.open("file://" + complete_file_path)

    def check_file(self, file_path: str):
        complete_file_path = os.path.join(self.root_path, file_path)
        return os.path.isfile(complete_file_path)

    def update_workspace(self, changes: dict[str, typing.Any]):
        abstra_json = self.__get_abstra_json()
        abstra_json.workspace.update(changes)
        self.persist(abstra_json)
        return abstra_json.workspace

    # Jobs CLRUD

    def get_jobs(self):
        abstra_json = self.__get_abstra_json()
        return abstra_json.jobs

    def get_job(self, identifier: str):
        abstra_json = self.__get_abstra_json()
        for job in abstra_json.jobs:
            if identifier == job.identifier:
                return job
        return None

    def create_job(self):
        abstra_json = self.__get_abstra_json()

        file_name = f"new_job_{random_id()}"
        file_path = f"{file_name}.py"

        with open(os.path.join(self.root_path, file_path), "w") as f:
            f.write("print('Hello World')")

        job = JobJSON(
            file=file_path,
            identifier=random_id(),
            schedule="* * * * *",
            title="Untitled Job",
        )

        abstra_json.jobs.append(job)

        self.persist(abstra_json)

        return job

    def update_job(self, identifier: str, changes: dict[str, typing.Any]):
        abstra_json = self.__get_abstra_json()
        jobs = [j for j in abstra_json.jobs if j.identifier == identifier]
        if len(jobs) == 0:
            raise Exception("Job not found for identifier ", identifier)
        job = jobs[0]
        job.update(changes)

        self.persist(abstra_json)

        return job

    def delete_job(self, identifier: str):
        abstra_json = self.__get_abstra_json()
        jobs = abstra_json.jobs
        abstra_json.jobs = [j for j in jobs if j.identifier != identifier]

        self.persist(abstra_json)



import inspect

from ...session import LiveSession
from ...contract import forms_contract

from abstra.forms.debug_utils import Frames, make_debug_data, CloseDTO

"""
class Connection:
    url_params: dict = {}

    def close(self, dto: CloseDTO):
        pass
        
    def send(self, data: dict, frames: Frames = None):
        pass
        
    def receive(self, path: str = "") -> any:
        pass

"""


class MessageBroker:
    session: LiveSession
    url_params: dict = {}

    def __init__(self, session: LiveSession):
        self.session = session
        self.__wait_start()

    def __wait_start(self):
        type = None
        while type != "start":
            type, data = self.session.recv()
        self.url_params = data["params"]

    def __browser_msg_handler__log_only(self, data: dict):
        # log data
        pass

    def __browser_msg_handler(self) -> dict:
        type, data = self.session.recv()
        if type in ["heartbeat", "executed-by", "metadata", "browser:try-disconnect"]:
            self.__browser_msg_handler__log_only(data)
            return self.__browser_msg_handler()
        return data

    # Connection interface
    def send(self, data: dict, frames: Frames = None):
        if self.session.is_preview:
            debug = make_debug_data(frames or inspect.stack())
            data.update(debug)
        self.session.send(forms_contract.GenericMessage(data))

    def receive(self, path: str = ""):
        data = self.__browser_msg_handler()
        if not path:
            return data
        return data.get(path)

    def close(self, dto: CloseDTO):
        self.send(
            {
                "type": "program:end",
                "exitCode": dto.exit_code,
                "exception": dto.exception,
            },
            frames=dto.frames,
        )
        self.session.close()



from ..apis import main as api_main
from . import utils as utils


def list_workspace_forms():
    query = """
        query GetForms {
            forms {
                path
                title
                script {
                    enabled
                }
            }
        }
    """
    return api_main.hf_hasura_runner(query).get("forms", [])


def add_workspace_form(data):
    _, workspace_id, _ = api_main.get_auth_info()
    form_data = {
        "title": data["name"],
        "workspace_id": workspace_id,
        "script": {
            "data": {
                "workspace_id": workspace_id,
                "enabled": data.get("enabled", True),
                "code": data["code"],
                "name": data["name"],
            }
        },
    }

    data.pop("name", None)
    data.pop("code", None)
    data.pop("enabled", None)
    form_data.update(data)

    query = """
        mutation InsertForm($form_data: [forms_insert_input!]!) {
            insert_forms(
                objects: $form_data
            ) {
                returning {
                    path
                    title
                }
            }
        }
    """

    return (
        api_main.hf_hasura_runner(query, {"form_data": form_data})
        .get("insert_forms", {})
        .get("returning", {})[0]
    )


def update_workspace_form(path, data):
    form_data = data.copy()
    script_data = {}

    name = form_data.pop("name", None)
    if name:
        form_data["title"] = name
        script_data["name"] = name

    code = form_data.pop("code", None)
    if code:
        script_data["code"] = code

    enabled = form_data.pop("enabled", None)
    if enabled is not None:
        script_data["enabled"] = enabled

    request_data = {"path": path, "form_data": form_data, "script_data": script_data}
    update_query = """
        mutation UpdateForm($path: String!, $form_data: forms_set_input, $script_data: scripts_set_input = {}) {
            update_forms(where: {path: {_eq: $path}}, _set: $form_data) {
                returning {
                    id
                    path
                    title
                }
            }
            update_scripts(where: {form: {path: {_eq: $path}}}, _set: $script_data) {
                returning {
                    name
                }
            }
        }
    """
    return api_main.hf_hasura_runner(update_query, request_data)


def upsert_workspace_form(data):
    path = data["path"]

    query = """
        query FindForm($path: String!) {
            forms(where: {path: {_eq: $path}}) {
                path
            }
        }
    """

    forms = api_main.hf_hasura_runner(query, {"path": path}).get("forms")
    if len(forms):
        return update_workspace_form(path, data)
    else:
        return add_workspace_form(data)


def delete_workspace_form(path):
    query = """
        mutation DeleteForm($path: String!) {
            delete_forms(where: {path: {_eq: $path}}) {
                returning {
                    id
                    path
                    title
                }
            }
        }
    """

    return api_main.hf_hasura_runner(query, {"path": path})


def list_logs(limit, offset):
    query = """
        query GetFormLogs($limit: Int, $offset: Int) {
            forms {
                id
                logs(offset: $offset, limit: $limit, order_by: {created_at: desc}) {
                    id
                    form_id
                    created_at
                    search_term
                    execution_type
                    messages
                }
            }
        }
    """

    forms = api_main.hf_hasura_runner(query).get("forms", [])
    logs = utils.flat_items_logs(forms, path_or_id="id")

    return {"logs": utils.sampling(logs, limit, offset)}


def list_logs_by_id(id, limit, offset):
    query = """
        query GetFormsLogs($limit: Int, $offset: Int, $id: uuid) {
            forms(where: {id: {_eq: $id}}) {
                id
                logs(offset: $offset, limit: $limit, order_by: {created_at: desc}) {
                    id
                    created_at
                    form_id
                    search_term
                    execution_type
                    messages
                }
            }
        }
    """
    forms = api_main.hf_hasura_runner(
        query, {"id": id, "limit": limit, "offset": offset}
    ).get("forms")
    if forms:
        return {"logs": utils.flatten_list([form.get("logs") for form in forms])}
    return {"logs": []}



import inspect, os, flask, werkzeug.exceptions as wz_ex
import abstra_statics

dist_folder = os.path.join(os.path.dirname(inspect.getfile(abstra_statics)), "dist")


def send_from_dist(filename: str, fallback: str):
    try:
        return flask.send_from_directory(dist_folder, filename)
    except wz_ex.NotFound as e:
        return flask.send_from_directory(dist_folder, fallback)



from .session import LiveSession
from .contract import common
from .utils import decode_jwt


class AuthResponse:
    """The response from the authentication process

    Attributes:
      email (str): The email address of the user
    """

    def __init__(self, email: str):
        self.email = email


def get_user(session: LiveSession):
    session.send(common.AuthRequireInfoMessage())
    while True:
        type, data = session.recv()
        if type != "auth:saved-jwt":
            continue

        jwt_claims = decode_jwt(data["jwt"])
        if not jwt_claims:
            session.send(common.AuthInvalidJWTMessage())
            continue

        session.send(common.AuthValidJWTMessage())
        return AuthResponse(jwt_claims["email"])
